import functools
import re
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, TypeVar, Union

import dotenv
import typer
import yaml
from pysh import cd, sh
from typing_extensions import Concatenate, ParamSpec, TypeAlias

SERVICE_PATHS_ARG = typer.Argument(
    None,
    help="Paths to service directories or root directories that contain multiple services. Current working directory by default",
    dir_okay=True,
    file_okay=False,
    exists=True,
    show_default=False,
)
VERBOSE_ARG = typer.Option(False, "-v", "--verbose", help="Print debugging output")
DOTENV_SECTION_SEPARATOR = "\n# =======================================\n"

ENV_VARS = {
    # Postgres
    "POSTGRES_HOST": '"localhost"',
    "POSTGRES_PORT": 5432,
    "POSTGRES_USER": "postgres",
    "POSTGRES_PASSWORD": "postgres",
    # Rabbit
    "RABBITMQ_HOST": '"localhost"',
    "RABBITMQ_PORT": 5672,
    "RABBITMQ_LOGIN": "guest",
    "RABBITMQ_PASSWORD": "guest",
    "RABBITMQ_VHOST": '"/"',
    "RABBITMQ_SSL": '"false"',
    # Etc
    "DEBUG": True,
    "CURRENT_ENV": "dev",
    "LOG_LEVEL": "INFO",
}

P = ParamSpec("P")
R = TypeVar("R")
PATHS: TypeAlias = Union[List[Path], None]

RE_PYTHON_VERSION = re.compile(r"(?P<major>\d+)\.(?P<minor>\d+)(\.(?P<bugfix>\d))?.*")


def parse_python_version(raw_python_version: str) -> Optional[Tuple[int, int]]:
    python_version = clean_python_version(raw_python_version)
    match = RE_PYTHON_VERSION.match(python_version)
    if match:
        return int(match["major"]), int(match["minor"])


def clean_python_version(version: str) -> str:
    return version.strip(" \n\t~^*><=")


def add_default_service_path(function: Callable[Concatenate[List[Path], P], R]) -> Callable[Concatenate[PATHS, P], R]:
    @functools.wraps(function)
    def wrapper(service_paths: PATHS, *args: P.args, **kwargs: P.kwargs) -> R:
        if not service_paths:
            service_paths = [Path.cwd()]
        return function(service_paths, *args, **kwargs)

    return wrapper


@dataclass(frozen=True)
class Service:
    dir: Path
    yaml_config: Union[Dict[str, Any], None]
    dotenv_path: Path
    dotenv: Dict[str, Union[str, None]]
    dotenv_example: Dict[str, Union[str, None]]
    dotenv_example_original_source: str


def get_service(dir: Path):
    dir = dir.absolute()
    return Service(
        dir,
        yaml.safe_load(safely_read_text(dir / ".helm/values.yaml")),
        dir / "settings/.env",
        dotenv.dotenv_values(dir / "settings/.env"),
        dotenv.dotenv_values(dir / "settings/.env.example"),
        safely_read_text(dir / "settings/.env.example"),
    )


def gather_services(paths: List[Path]) -> Dict[str, Service]:
    service_dirs: List[Path] = []
    for path in paths:
        path = path.resolve()

        if is_service_dir(path):
            service_dirs.append(path)
        else:
            service_dirs.extend(unpack_root_path(path))

    return {dir.name: get_service(dir) for dir in service_dirs}


def safely_read_text(path: Path) -> str:
    return path.read_text() if path.is_file() else ""


def is_service_dir(path: Path) -> bool:
    return path.is_dir() and (path / "settings/.env.example").exists()


def unpack_root_path(path: Path) -> List[Path]:
    return [p for p in path.iterdir() if is_service_dir(p)]


def save_dotenv_file(service: Service) -> None:
    """I save the dotenv file while preserving the comments and the order of entries"""
    new_lines: List[str] = []
    dotenv_items = service.dotenv.copy()
    for line in service.dotenv_example_original_source.splitlines():
        line = line.strip()
        if not line:
            new_lines.append("")
        elif line.startswith("#"):
            new_lines.append(line)
        elif "=" in line:
            key, _ = line.split("=", 1)
            key = key.strip()
            if key in dotenv_items:
                new_lines.append(f"{key}={dotenv_items.pop(key) or ''}")

    new_lines.append(DOTENV_SECTION_SEPARATOR + "# Env vars not present in .env.example" + DOTENV_SECTION_SEPARATOR)
    for key, value in dotenv_items.items():
        new_lines.append(f"{key}={value or ''}")

    service.dotenv_path.write_text("\n".join(new_lines))


def sh_with_log(cmd: str, prefix: str = "\n", suffix: str = "\n", capture: bool = False):
    typer.echo(f"{prefix}>>> {cmd}")
    res = sh(cmd, capture=capture)
    typer.echo(f"{suffix}")
    return res


@contextmanager
def cd_with_log(directory: "Path | str", prefix: str = "") -> Iterator[Path]:
    directory = Path(directory)
    if Path.cwd() == directory.resolve():
        yield directory
        return

    typer.echo(f"{prefix}>>> cd {directory}")
    with cd(directory) as path:
        yield path
    typer.echo(f"{prefix}>>> cd -")



# Taken from poetry project: https://github.com/python-poetry

import re
from dataclasses import dataclass
from pathlib import Path
from typing import List, Union

import tomli as toml


@dataclass
class DependencySpec:
    name: str
    path: Union[str, None] = None
    version: Union[str, None] = None
    extras: Union[List[str], None] = None


def _parse_dependency_specification_simple(requirement: str) -> Union[DependencySpec, None]:
    extras: List[str] = []
    pair = re.sub("^([^@=: ]+)(?:@|==|(?<![<>~!])=|:| )(.*)$", "\\1 \\2", requirement)
    pair = pair.strip()

    require = DependencySpec("")

    if " " in pair:
        name, version = pair.split(" ", 2)
        extras_m = re.search(r"\[([\w\d,-_]+)\]$", name)
        if extras_m:
            extras = [e.strip() for e in extras_m.group(1).split(",")]
            name, _ = name.split("[")

        require.name = name
        require.version = version
    else:
        m = re.match(r"^([^><=!: ]+)((?:>=|<=|>|<|!=|~=|~|\^).*)$", requirement.strip())
        if m:
            name, constraint = m.group(1), m.group(2)
            extras_m = re.search(r"\[([\w\d,-_]+)\]$", name)
            if extras_m:
                extras = [e.strip() for e in extras_m.group(1).split(",")]
                name, _ = name.split("[")

            require.name = name
            require.version = constraint
        else:
            extras_m = re.search(r"\[([\w\d,-_]+)\]$", pair)
            if extras_m:
                extras = [e.strip() for e in extras_m.group(1).split(",")]
                pair, _ = pair.split("[")

            require.name = pair

    if extras:
        require.extras = extras

    return require


def _parse_dependency_specification_path(requirement):
    path = Path(requirement)
    if path.exists() and path.is_dir():
        pyproject = path / "pyproject.toml"
        if pyproject.exists():
            package_name = toml.loads(pyproject.read_text())["tool"]["poetry"]["name"]
            return DependencySpec(name=package_name, path=str(path))


def parse_dependency_specification(requirement: str) -> DependencySpec:
    requirement = requirement.strip()

    extras = []
    extras_m = re.search(r"\[([\w\d,-_ ]+)\]$", requirement)
    if extras_m:
        extras = [e.strip() for e in extras_m.group(1).split(",")]
        requirement, _ = requirement.split("[")

    specification = _parse_dependency_specification_path(requirement) or _parse_dependency_specification_simple(
        requirement
    )

    if specification:
        if extras and not specification.extras:
            specification.extras = extras
        return specification

    raise ValueError(f"Invalid dependency specification: {requirement}")



from typing import List

import typer
from pysh import which

from stb import config, db, graph, run, update, use
from stb.__version__ import __version__

app = typer.Typer(
    name="stb",
    add_completion=False,
    help="Stanislav's Toolbox (stb) helps you manage your local microservices. Specifically, it can set them up for you, create/delete/migrate databases, manage service ports and .env files, and much more",
)

app.add_typer(db.app, name="db")
app.add_typer(update.app)
app.add_typer(config.config_app)
app.add_typer(graph.app)


def version_callback(value: bool):
    if value:
        typer.echo(f"Stanislav's Toolbox {__version__}")
        raise typer.Exit()


@app.command(name="use")
def use_(
    requirements: List[str] = typer.Argument(
        ...,
        help="Package versions to use. For example, `my_package` or `my_package==3.1.4` or `~/path/to/my_package`",
    ),
    editable: bool = typer.Option(
        False,
        help="Install the package(s) in editable mode. Useful when you want to develop the package and use it in the current project at the same time",
    ),
    fix: bool = typer.Option(
        False,
        help="Fix the broken version of the package in the current project as well. Useful when dependency resolution doesn't work properly",
    ),
) -> None:
    """Switches the version of a company package in the current project. For example, `stb use my_package 0.1.0` or `stb use my_package ~/package`"""
    return use.use_packages(requirements, editable, fix)


def run_(
    services: List[str] = typer.Argument(
        ...,
        help="The select services to checkout and run together at the same time",
    ),
) -> None:
    return run.run_services(set(services))


if which("concurrently"):
    run_ = app.command(name="run")(run_)

try:
    from stb import setup

    @app.command(name="setup")
    def setup_(
        services: List[str] = typer.Argument(
            ...,
            help="Names of services to setup. For example, use 'my_company/backend/oatmeal' to setup oatmeal service in the backend namespace",
        ),
        skip_existing: bool = typer.Option(True, help="Automatically skip existing services"),
        update_env: bool = typer.Option(True, help="Generate .env settings files"),
        update_ports: bool = typer.Option(True, help="Add dependended services' ports to .env settings files"),
        setup_poetry_env: bool = typer.Option(True, help="Pick the correct python version and install dependencies"),
    ) -> None:
        """Does the initial localhost setup of microservices. Downloads, configures .env, inits submodules, installs the correct pyenv environment, creates the correct poetry environment, and installs dependencies"""
        return setup.setup_services(
            services,
            skip_existing,
            update_env=update_env,
            update_ports=update_ports,
            setup_poetry_env=setup_poetry_env,
        )

except ImportError:
    pass


@app.callback()
def main(
    version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=True),
):
    ...


if __name__ == "__main__":
    app()



import itertools
from pathlib import Path
from typing import Any, List

import tomli as toml
import typer
from pysh import sh, which

from stb.config import CONFIG
from stb.utils.common import sh_with_log

from .utils.dependency_parser import parse_dependency_specification

PYENV_INSTALLED = which("pyenv")


def use_packages(requirements: List[str], editable: bool = False, fix: bool = False) -> None:
    editable_arg = "--editable" if editable else ""
    pyproject_path = Path("pyproject.toml")
    pyproject = toml.loads(pyproject_path.read_text())
    dependencies = pyproject["tool"]["poetry"]["dependencies"]
    if not pyproject_path.exists():
        raise FileNotFoundError("No pyproject.toml found in the current directory. It is required by 'use' to work.")
    existing_requirements: List[str] = []
    version_based_requirements: List[str] = []
    path_based_requirements: List[str] = []

    for requirement in requirements:
        spec = parse_dependency_specification(requirement)
        if spec.name in dependencies:
            existing_requirements.append(spec.name)

        if not spec.path and editable:
            raise typer.BadParameter("Editable mode is only supported for local packages.")

        old_value = extract_package_info_from_pyproject(spec.name, dependencies)
        version = spec.version or old_value.get("version", "latest")
        extras = old_value.get("extras", [])
        extras = list(itertools.chain.from_iterable([e.split() for e in extras]))

        if spec.path:
            extras_arg = " ".join(["-E " + e for e in extras])
            path_based_requirements.append(requirement + " " + extras_arg)
        else:
            if not "pypi_source" in CONFIG:
                raise typer.BadParameter(
                    "You must set the pypi_source in the config file before you can use this command"
                )

            extras_arg = f'[{",".join(extras)}]' if extras else ""
            version_based_requirements.append(f'"{spec.name}{extras_arg}@{version}"')

    if fix:
        sh_with_log(f"poetry remove {' '.join(existing_requirements)}")

    for requirement in path_based_requirements:
        sh_with_log(f"poetry add {requirement} {editable_arg}")

    if version_based_requirements:
        formatted_version_based_requirements = " ".join(version_based_requirements)
        sh_with_log(f"poetry add {formatted_version_based_requirements} --source={CONFIG['pypi_source']}")


def extract_package_info_from_pyproject(package_name: str, dependencies: "dict[str, str | dict]") -> "dict[str, Any]":
    if package_name in dependencies:
        old_value = dependencies[package_name]

        # oatmeal = "3.8.3"
        if not isinstance(old_value, dict):
            return {"version": old_value}
        # oatmeal = {version="3.8.3", extras=["server"], source="oatmeal_pypi"}
        elif "version" in old_value:
            return old_value
        # oatmeal = {path = "/path/to/local/oatmeal/package", extras = ["server"], develop = true}
        elif "path" in old_value:
            return old_value
        else:
            raise NotImplementedError(f"Unexpected dependency value for {package_name}: {dependencies[package_name]}")
    else:
        return {}



#!/usr/bin/env python3

import enum
import multiprocessing
from contextlib import suppress
from pathlib import Path
from typing import List, cast

import rich
import typer
from pysh import env

from .utils.common import SERVICE_PATHS_ARG, Service, add_default_service_path, cd_with_log, get_service, sh_with_log


def old_parallel_flag_deprecation_callback(value: bool):
    if value:
        rich.print(
            "[bold red]The flag '-p' for 'stb db' is deprecated and will be removed in the future. Its behavior is now the default[/bold red]\n",
        )


app = typer.Typer(
    name="migrator",
    help="creates/upgrades/drops dbs for microservices",
)
REQUIRED_DOTENV_KEYS = "POSTGRES_PASSWORD", "POSTGRES_USER"
OLD_PARALLEL_MIGRATIONS_ARG = typer.Option(
    False,
    "-p",
    "--parallel",
    help="Deprecated. This behavior is now the default. If you want to disable it, use --no-parallel.",
    callback=old_parallel_flag_deprecation_callback,
)
NO_PARALLEL_MIGRATIONS_ARG = typer.Option(False, "-P", "--no-parallel", help="Do not run migrations in parallel")
FORCE_DROP_ARG = typer.Option(
    False,
    "-f",
    "--force",
    help="Force the drop of the databases. Helpful if other clients are connected to any of the databases. Use with caution.",
)


class Choices(str, enum.Enum):
    upgrade = "upgrade"
    create = "create"
    drop = "drop"


@app.command()
@add_default_service_path
def upgrade(
    service_paths: List[Path] = SERVICE_PATHS_ARG,
    no_parallel_migrations: bool = NO_PARALLEL_MIGRATIONS_ARG,
    old_parallel_migrations: bool = OLD_PARALLEL_MIGRATIONS_ARG,
):
    """Upgrade database migrations"""
    run_on_several_services(service_paths, Choices.upgrade, not no_parallel_migrations)


@app.command()
@add_default_service_path
def create(
    service_paths: List[Path] = SERVICE_PATHS_ARG,
    no_parallel_migrations: bool = NO_PARALLEL_MIGRATIONS_ARG,
    old_parallel_migrations: bool = OLD_PARALLEL_MIGRATIONS_ARG,
):
    """Create databases and upgrade their migrations"""
    run_on_several_services(service_paths, Choices.create, not no_parallel_migrations)


@app.command()
@add_default_service_path
def drop(
    service_paths: List[Path] = SERVICE_PATHS_ARG,
    force: bool = FORCE_DROP_ARG,
):
    """Drop databases"""
    run_on_several_services(service_paths, Choices.drop, force)


@app.command()
@add_default_service_path
def reset(
    service_paths: List[Path] = SERVICE_PATHS_ARG,
    no_parallel_migrations: bool = NO_PARALLEL_MIGRATIONS_ARG,
    force: bool = FORCE_DROP_ARG,
    old_parallel_migrations: bool = OLD_PARALLEL_MIGRATIONS_ARG,
):
    """Drop databases, recreate them, and then upgrade their migrations"""
    for service in service_paths:
        with suppress(Exception):
            run_on_single_service(service, Choices.drop, not no_parallel_migrations, force)
            run_on_single_service(service, Choices.create, not no_parallel_migrations)


def run_on_several_services(
    service_paths: List[Path],
    choice: Choices,
    parallel_migrations: bool = False,
    force_drop: bool = False,
):
    for service in service_paths:
        with suppress(Exception):
            run_on_single_service(service, choice, parallel_migrations, force_drop)


def run_on_single_service(
    service_path: Path,
    command: Choices,
    parallel_migrations: bool = False,
    force_drop: bool = False,
) -> None:
    service = get_service(service_path)

    for field in REQUIRED_DOTENV_KEYS:
        if field not in service.dotenv:
            err = f"{field} field is required for the correct functioning of stb db but it was not filled out in {service.dotenv_path}"
            typer.echo(err, err=True)
            raise LookupError(err)
    if not "POSTGRES_PORT" in service.dotenv:
        service.dotenv["POSTGRES_PORT"] = "5432"

    aerich_apps = find_aerich_apps(service)
    postgres_dbs = {v for k, v in service.dotenv.items() if k.startswith("POSTGRES_DB")}
    postgres_user = service.dotenv["POSTGRES_USER"]
    postgres_password = cast(str, service.dotenv["POSTGRES_PASSWORD"])
    postgres_port = service.dotenv["POSTGRES_PORT"]

    with cd_with_log(service.dir), env(PGPASSWORD=postgres_password):
        for db in aerich_apps | postgres_dbs:
            if command == Choices.create:
                sh_with_log(f"createdb -h localhost -p {postgres_port} -U {postgres_user} {db}", "", "")
            elif command == Choices.drop:
                sh_with_log(
                    f"dropdb {'-f' if force_drop else ''} -h localhost -p {postgres_port} -U {postgres_user} {db}",
                    "",
                    "",
                )

        if command in {Choices.create, Choices.upgrade}:
            if "aerich" in aerich_apps:
                aerich_apps.remove("aerich")
            sh_with_log("poetry run aerich upgrade")
            commands_to_run = [f"poetry run aerich --app {db} upgrade" for db in aerich_apps]

            if parallel_migrations:
                with multiprocessing.Pool() as pool:
                    pool.map(sh_with_log, commands_to_run)
            else:
                for cmd in commands_to_run:
                    sh_with_log(cmd)


def find_aerich_apps(service: Service) -> "set[str]":
    migrations_dir = service.dir / "migrations"
    if migrations_dir.is_dir():
        return {d.name for d in migrations_dir.iterdir() if d.is_dir()}
    else:
        return set()


if __name__ == "__main__":
    app()



import contextlib
import functools
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional, ParamSpec, Tuple, TypeVar

import keyring
import keyring.errors
import tomlkit
import typer
from platformdirs import user_config_dir

from .utils.common import sh_with_log

STB_APP_CONFIG_NAME = "stb"
STB_APP_AUTHOR_NAME = "ovsyanka83"
APP_TOKEN_NAME = "stb_app_token"

set_app = typer.Typer(
    name="set",
    help="Set config values in the stb config file",
)
get_app = typer.Typer(
    name="get",
    help="Get config values from the stb config file",
)

config_app = typer.Typer(
    name="config",
    help="Set config values in the stb config file",
)

config_app.add_typer(set_app)
config_app.add_typer(get_app)
P = ParamSpec("P")
T = TypeVar("T")


@dataclass(init=False)
class Config:
    def __init__(self) -> None:
        self.config_dir = Path(user_config_dir(STB_APP_CONFIG_NAME, STB_APP_AUTHOR_NAME))
        self.config_dir.mkdir(parents=True, exist_ok=True)
        self.config_file = self.config_dir / "cfg.toml"
        self.doc = tomlkit.loads(self.config_file.read_text()) if self.config_file.exists() else tomlkit.document()

    def __setitem__(self, key: str, value: Any) -> None:
        self.doc.__setitem__(key, value)

    def __getitem__(self, key: str) -> str:
        if key == "gitlab_api_token":
            return self.get_api_token()
        return str(self.doc.__getitem__(key))

    def __contains__(self, key: str) -> bool:
        if key == "gitlab_api_token":
            with contextlib.suppress(ValueError, typer.BadParameter):
                self.get_api_token()
                return True
        return self.doc.__contains__(key)

    def get_api_token(self) -> str:
        """Retrieves the gitLab api token from the keychain system service"""
        if "gitlab_api_token_name" not in self:
            raise typer.BadParameter("gitlab_api_token_name is not set yet")

        api_token_name = self["gitlab_api_token_name"]

        try:
            token = keyring.get_password(APP_TOKEN_NAME, api_token_name)
        except (RuntimeError, keyring.errors.KeyringError):
            raise ValueError(f"Unable to retrieve the password for {api_token_name} from the keyring.")

        if token is None:
            raise typer.BadParameter("gitlab_api_token is not set yet.")
        return token

    def set_api_token(self, name: str, value: str) -> None:
        try:
            keyring.set_password(APP_TOKEN_NAME, name, value)
        except (RuntimeError, keyring.errors.KeyringError) as e:
            raise ValueError(f"Unable to store the token for {name} in the keyring: {e}")

    def save(self, path: Optional[Path] = None) -> None:
        if path is None:
            path = self.config_file
        path.write_text(tomlkit.dumps(self.doc))

    def requires(self, first_key: str, *keys: str):
        def decorator(func: Callable[P, T]) -> Callable[P, T]:
            @functools.wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                self.check_keys_have_been_set(first_key, *keys)
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def check_keys_have_been_set(self, *keys: str) -> None:
        missing_keys = [key for key in keys if key not in self]
        if missing_keys:
            raise typer.BadParameter(
                "\n".join(
                    f"You must set the '{key}' parameter in the config file before you can use this command."
                    for key in missing_keys
                )
                + "\n\nSee `stb config set --help for more information`"
            )


CONFIG = Config()

RE_GIT_URL = re.compile(r"^git@[a-zA-Z0-9._-]+$")


def make_command(
    entry_name: str,
    help: str = "",
    regex: Optional[re.Pattern] = None,
    set_callback: Callable[[str], None] = lambda _: None,
) -> Tuple[Callable[[], Any], Callable[[Any], None]]:
    def get_command():
        if entry_name not in CONFIG:
            raise typer.BadParameter(f"{entry_name} is not set yet")
        return typer.echo(CONFIG[entry_name])

    def set_command(value: str = typer.Argument(..., help=help, show_default=False)) -> None:
        set_callback(value)
        if regex is not None and regex.match(value) is None:
            raise typer.BadParameter(f"{value} does not match the expected format")
        CONFIG[entry_name] = value
        CONFIG.save()

    get_command.__doc__ = f"Get the {help}"
    set_command.__doc__ = f"Set the {help}"

    return get_app.command(entry_name)(get_command), set_app.command(entry_name)(set_command)


get_git_url, set_git_url = make_command(
    "git_url", "git url for setting up local services in the following format: 'git@github.com'", RE_GIT_URL
)

get_pypi_source, set_pypi_source = make_command(
    "pypi_source", "internal pypi source name that is used for downloading internal packages"
)

get_pypi_registry_id, set_pypi_registry_id = make_command(
    "pypi_registry_id",
    "id of the package whose registry is used for downloading internal packages",
    re.compile(r"^\d+$"),
)


@get_app.command("gitlab_api_token")
def get_gitlab_api_token() -> None:
    """Get the gitlab api token for setting up local services"""
    api_token = CONFIG.get_api_token()
    api_token_name = CONFIG["gitlab_api_token_name"]

    typer.echo(f"{api_token_name} {api_token}")


@set_app.command("gitlab_api_token")
def set_gitlab_api_token(
    token_name: str = typer.Argument(..., help="token name"),
    token: str = typer.Argument(..., help="token itself"),
) -> None:
    """Set the gitlab api token name for setting up local services"""
    CONFIG["gitlab_api_token_name"] = token_name
    CONFIG.save()
    CONFIG.set_api_token(token_name, token)


@CONFIG.requires("pypi_source", "pypi_registry_id", "gitlab_api_token", "git_url")
@config_app.command()
def poetry() -> None:
    """Setup poetry config based on exiting stb config"""

    api_url = get_gitlab_api_url()

    sh_with_log(
        f"poetry config repositories.{CONFIG['pypi_source']} {api_url}/projects/{CONFIG['pypi_registry_id']}/packages/pypi"
    )
    sh_with_log(
        f"poetry config http-basic.{CONFIG['pypi_source']} {CONFIG['gitlab_api_token_name']} {CONFIG['gitlab_api_token']}"
    )


def get_gitlab_api_url() -> str:
    git_host = CONFIG["git_url"].split("@")[1]
    return f"https://{git_host}/api/v4"


# TODO: Delete after everyone has updated to >=3.0.0
if "git_url" in CONFIG and ":" in CONFIG["git_url"]:
    CONFIG["git_url"] = CONFIG["git_url"].split(":")[0]
    CONFIG.save()

# TODO: Delete after everyone has updated to >=4.5.0
if "gitlab_api_token" in CONFIG.doc:
    if "gitlab_api_token_name" in CONFIG.doc:
        try:
            CONFIG.get_api_token()
        except typer.BadParameter:
            typer.echo("Adding gitlab api token into keyring", err=True)
            try:
                CONFIG.set_api_token(CONFIG.doc["gitlab_api_token_name"], CONFIG.doc["gitlab_api_token"])
            except Exception as e:
                typer.echo(f"Something went wrong, please set the token again. Error message:\n{e}", err=True)

    typer.echo("Removing gitlab_api_token from config", err=True)
    CONFIG.doc.remove("gitlab_api_token")
    CONFIG.save()



import contextlib
import datetime
import io
import json
import subprocess
from typing import List

import gitlab
import tomli
import typer
from pysh import which

from .config import CONFIG, get_gitlab_api_url

app = typer.Typer(
    name="graph",
    help="graphs the dependencies of microservices using various backends",
)
ignore_packages = typer.Option(
    None,
    "-i",
    help="The packages to omit from output even if they are in the registry.",
)

REPLACEMENTS = {
    "xchange-rates": "exchange-rates",
    "mailbox": "mailboxes",
}

GRAPHVIZ_INPUT = """
digraph prof {{
	ratio = fill;
	node [style=filled];
    {}
}}
"""

today = datetime.datetime.today().strftime("%y%m%d_%H%M%S")

if which("dot"):

    @app.command()
    def graphviz(
        services: List[str],
        ignore_packages: list[str] = ignore_packages,
    ):
        """Graphs the dependencies of microservices using graphviz"""
        with contextlib.redirect_stdout(io.StringIO()):
            deps = json_(services, ignore_packages)
        stuff = "\n".join([f"{k.replace('-', '_')} -> {dep.replace('-', '_')}" for k, v in deps.items() for dep in v])
        subprocess.run(f"dot -Tsvg", shell=True, text=True, input=GRAPHVIZ_INPUT.format(stuff))


@CONFIG.requires("gitlab_api_token", "git_url", "pypi_registry_id")
@app.command(name="json")
def json_(
    services: List[str],
    ignore_packages: list[str] = ignore_packages,
):
    gl = gitlab.Gitlab(url=get_gitlab_api_url().removesuffix("/api/v4"), private_token=CONFIG["gitlab_api_token"])
    projects = get_projects(gl, services)
    dep_mapping = {}
    project_with_registry = gl.projects.get(CONFIG["pypi_registry_id"])
    packages = project_with_registry.packages.list(all=True)
    package_names_in_registry = {package.name for package in packages}

    for pyproject in get_pyproject_tomls(projects):
        service_name = get_service_name(pyproject)
        if not service_name:
            continue

        direct_dependencies = get_direct_dependencies(pyproject)
        direct_registry_dependencies = [
            dep for dep in direct_dependencies if dep in package_names_in_registry and not dep in ignore_packages
        ]
        if service_name in direct_registry_dependencies:
            direct_registry_dependencies.remove(service_name)
        dep_mapping[service_name] = direct_registry_dependencies
    dep_mapping = {k: v for k, v in dep_mapping.items() if dep_is_in_mapping(dep_mapping, k)}
    typer.echo(json.dumps(dep_mapping, indent=4, ensure_ascii=False))
    return dep_mapping


def dep_is_in_mapping(dep_mapping: dict[str, list[str]], searched_service: str):
    if dep_mapping.get(searched_service):
        return True
    for dependencies in dep_mapping.values():
        for dependency in dependencies:
            if dependency == searched_service:
                return True
    return False


def get_service_name(pyproject: dict) -> str:
    service_name = "-".join(pyproject["tool"]["poetry"]["name"].replace("_", "-").split("-")[:-1]).strip()
    if service_name in REPLACEMENTS:
        return REPLACEMENTS[service_name]
    return service_name


def get_direct_dependencies(pyproject):
    return [dep.replace("_", "-") for dep in pyproject["tool"]["poetry"]["dependencies"].keys()]


def get_pyproject_tomls(projects: list):
    for project in projects:
        with contextlib.suppress(gitlab.GitlabGetError):
            possible_pyproject_ids = [
                i["id"] for i in project.repository_tree(get_all=True) if i["path"] == "pyproject.toml"
            ]
            if possible_pyproject_ids:
                yield tomli.loads(project.repository_raw_blob(possible_pyproject_ids[0]).decode("utf-8"))


def get_projects(gl: gitlab.Gitlab, repo_names: List[str]) -> list:
    all_projects = gl.projects.list(get_all=True)

    calculated_projects = []
    repo_names = ["".join(name.split()) for name in repo_names]
    for name in repo_names:
        if name.count("/") == 2:
            calculated_projects.append(name.split("/")[-1])

    for name in repo_names:
        # Remove all whitespace in case the user accidentally added some
        projects = [p for p in all_projects if p.path_with_namespace.startswith(name)]
        if not projects:
            raise ValueError(
                f"Failed to find any projects that start with '{name}'. Maybe you need to add a group/namespace or to fix a typo?"
            )
        calculated_projects.extend(projects)

    return calculated_projects
